/*
  This Code is to be Loaded onto the Arduino in the Free-Fall Object

  Code should be commented out for IMUs not used in the Free-Fall Object
*/


#include <ArduinoBLE.h>
#include <Arduino_BMI270_BMM150.h>
#include <Adafruit_ICM20X.h>
#include <Adafruit_ICM20649.h>
#include <Adafruit_ICM20948.h>
#include <Adafruit_ISM330DHCX.h>
#include <Adafruit_LSM6DSO32.h>
#include <Adafruit_Sensor.h>
#include <SensorFusion.h>

#define GRAVITY_METERS_SECOND_2 9.80665
#define BMI270_SAMPLERATE_DELAY_MS 15
#define SAMPLERATE_DELAY_MS 5

//
// Set the IMUs you are Using to True
//
bool useBmi270 = false;
bool useIcm20649 = false;
bool useIcm20948 = false;
bool useIsm330dhcx = false;
bool useLsm6dso32 = false;
//
// End of IMUs Selections
//


float bmi270Deltat, icm20649Deltat, icm20948Deltat, ism330dhcxDeltat, lsm6dso32Deltat;
float* quat;
float g_x, g_y, g_z;
unsigned long startTime;
unsigned long elapsedTime;
float bmi270ImuData[200][10];
float icm20649ImuData[200][10];
float icm20948ImuData[200][10];
float ism330dhcxImuData[200][10];
float lsm6dso32ImuData[200][10];
String imuSnapshot[202];


//
// Enter the Offset and Scale Values you Recorded for the IMUs You Have Calibrated
//
const float bmi270OffsetX = 0;
const float bmi270OffsetY = 0;
const float bmi270OffsetZ = 0;
const float bmi270ScaleX = GRAVITY_METERS_SECOND_2;
const float bmi270ScaleY = GRAVITY_METERS_SECOND_2;
const float bmi270ScaleZ = GRAVITY_METERS_SECOND_2;

const float icm20649OffsetX = 0;
const float icm20649OffsetY = 0;
const float icm20649OffsetZ = 0;
const float icm20649ScaleX = GRAVITY_METERS_SECOND_2;
const float icm20649ScaleY = GRAVITY_METERS_SECOND_2;
const float icm20649ScaleZ = GRAVITY_METERS_SECOND_2;

const float icm20948OffsetX = 0;
const float icm20948OffsetY = 0;
const float icm20948OffsetZ = 0;
const float icm20948ScaleX = GRAVITY_METERS_SECOND_2;
const float icm20948ScaleY = GRAVITY_METERS_SECOND_2;
const float icm20948ScaleZ = GRAVITY_METERS_SECOND_2;

const float ism330dhcxOffsetX = 0;
const float ism330dhcxOffsetY = 0;
const float ism330dhcxOffsetZ = 0;
const float ism330dhcxScaleX = GRAVITY_METERS_SECOND_2;
const float ism330dhcxScaleY = GRAVITY_METERS_SECOND_2;
const float ism330dhcxScaleZ = GRAVITY_METERS_SECOND_2;

const float lsm6dso32OffsetX = 0;
const float lsm6dso32OffsetY = 0;
const float lsm6dso32OffsetZ = 0;
const float lsm6dso32ScaleX = GRAVITY_METERS_SECOND_2;
const float lsm6dso32ScaleY = GRAVITY_METERS_SECOND_2;
const float lsm6dso32ScaleZ = GRAVITY_METERS_SECOND_2;
//
// End of Offset and Scale Values
//

SF bmi270Fusion;
SF icm20649Fusion;
SF icm20948Fusion;
SF ism330dhcxFusion;
SF lsm6dso32Fusion;

Adafruit_ICM20649 icm20649;
Adafruit_ICM20948 icm20948;
Adafruit_ISM330DHCX ism330dhcx;
Adafruit_LSM6DSO32 lsm6dso32;

BLEService freeFallObjectService("19B10000-E8F2-537E-4F6C-D104768A1214"); // Bluetooth® Low Energy LED Service
BLEByteCharacteristic activateIMUCharacteristic("0001", BLERead | BLEWrite); // Activate IMU Characteristic
BLEStringCharacteristic transferIMUDataCharacteristic("0002", BLERead | BLEIndicate, 255); // Transfer IMU Data Characteristic

void setup() {
  Serial.begin(115200);
  delay(5000);
  //while (!Serial);
  Wire.begin();

  // begin initialization
  if (!BLE.begin()) {
    Serial.println("starting Bluetooth® Low Energy module failed!");
    while (1);
  }

  //
  // Initialize the IMUs You are Using
  //
  if (useBmi270) {
    initializeBmi270IMU();
  }
  if (useIcm20649) {
    initializeIcm20649IMU();
  }
  if (useIcm20948) {
    initializeIcm20948IMU();
  }
  if (useIsm330dhcx) {
    initializeIsm330dhcxIMU();
  }
  if (useLsm6dso32) {
    initializeLsm6dso32IMU();
  }
  //
  // End of Intialization Calls
  //

  // set advertised local name and service UUID:
  BLE.setLocalName("Free-Fall Object");
  BLE.setAdvertisedService(freeFallObjectService);

  // add the characteristic to the service
  freeFallObjectService.addCharacteristic(activateIMUCharacteristic);
  freeFallObjectService.addCharacteristic(transferIMUDataCharacteristic);

  // add service
  BLE.addService(freeFallObjectService);

  // set the initial value for the characeristic:
  activateIMUCharacteristic.writeValue(0);

  // start advertising
  BLE.advertise();

  Serial.println("Free-Fall Object Peripheral");

  Wire.setClock(4000000);
}

void loop() {
  // listen for Bluetooth® Low Energy peripherals to connect:
  BLEDevice central = BLE.central();

  // if a central is connected to peripheral:
  if (central) {
    Serial.print("Connected to central: ");
    // print the central's MAC address:
    Serial.println(central.address());

    // while the central is still connected to peripheral:
    while (central.connected()) {
      // Allow Sensor Fusion Algorithm to Calibrate
      if (useBmi270) {
        sfBmi270Calibrate();
      }
      if (useIcm20649) {
        sfIcm20649Calibrate();
      }
      if (useIcm20948) {
        sfIcm20948Calibrate();
      }
      if (useIsm330dhcx) {
        sfIsm330dhcxCalibrate();
      }
      if (useLsm6dso32) {
        sfLsm6dso32Calibrate();
      }
      // if the remote device wrote to the characteristic,
      // use the value to control the LED:
      if (activateIMUCharacteristic.written()) {
        if (activateIMUCharacteristic.value()) {   // any value other than 0
          for (int i = 0; i < 200; i++) {
            // Record the start time
            startTime = millis();
            if (useBmi270) {
              recordBmi270IMUData(i);
            }
            if (useIcm20649) {
              recordIcm20649IMUData(i);
            }
            if (useIcm20948) {
              recordIcm20948IMUData(i);
            }
            if (useIsm330dhcx) {
              recordIsm330dhcxIMUData(i);
            }
            if (useLsm6dso32) {
              recordLsm6dso32IMUData(i);
            }
            // Calculate how long the loop took to execute
            elapsedTime = millis() - startTime;
            Serial.println(elapsedTime);

            // Delay for the remaining time to ensure 10ms per loop iteration, if necessary
            if ((useBmi270) && (elapsedTime < BMI270_SAMPLERATE_DELAY_MS)) {
              delay(BMI270_SAMPLERATE_DELAY_MS - elapsedTime);
            }
            if (!(useBmi270) && (elapsedTime < SAMPLERATE_DELAY_MS)) {
              delay(SAMPLERATE_DELAY_MS - elapsedTime);
            }
          }
          if (useBmi270) {
            bmi270ProcessData();
          }
          if (useIcm20649) {
            icm20649ProcessData();
          }
          if (useIcm20948) {
            icm20948ProcessData();
          }
          if (useIsm330dhcx) {
            ism330dhcxProcessData();
          }
          if (useLsm6dso32) {
            lsm6dso32ProcessData();
          }
          for (int i = 0; i < 202; i++) {
            imuSnapshot[i] = "";
          }
          for (int i = 0; i < 202; i++) {
            if (useBmi270) {
              bmi270SnapshotToString(i);
            }
            if (useIcm20649) {
              icm20649SnapshotToString(i);
            }
            if (useIcm20948) {
              icm20948SnapshotToString(i);
            }
            if (useIsm330dhcx) {
              ism330dhcxSnapshotToString(i);
            }
            if (useLsm6dso32) {
              lsm6dso32SnapshotToString(i);
            }
          }
          transferIMUData();
        }
      }
    }

    // when the central disconnects, print it out:
    Serial.print(F("Disconnected from central: "));
    Serial.println(central.address());
  }
}


//
// IMU Initialization Functions
//
void initializeBmi270IMU() {
  if (!IMU.begin(BOSCH_ACCELEROMETER_ONLY)) {
    Serial.println("Failed to initialize BMI270 IMU");
    while (1);
  }
  // Initialise BMI270 IMU Device
  Serial.print("BMI270 Accelerometer Sample Rate = ");
  Serial.print(IMU.accelerationSampleRate());
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Acceleration in Meters/Second2");
  Serial.println("X\tY\tZ");

  Serial.print("BMI270 Gyroscope sample rate = ");
  Serial.print(IMU.gyroscopeSampleRate());
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Gyroscope in Radians");
  Serial.println("X\tY\tZ");
}
void initializeIcm20649IMU() {
  if (!icm20649.begin_I2C()) {
    Serial.println("Failed to initialize ICM20649 IMU");
    while (1);
  }
  // Initialise IMU Device
  icm20649.setAccelRange(ICM20649_ACCEL_RANGE_4_G);
  icm20649.setAccelRateDivisor(4.6);
  Serial.print("ICM20649 Accelerometer Sample Rate = ");
  uint16_t accel_divisor = icm20649.getAccelRateDivisor();
  float accel_rate = 1125 / (1.0 + accel_divisor);
  Serial.print(accel_rate);
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Acceleration in Meters/Second2");
  Serial.println("X\tY\tZ");

  icm20649.setGyroRateDivisor(4.5);
  Serial.print("ICM20649 Gyroscope Sample Rate = ");
  uint8_t gyro_divisor = icm20649.getGyroRateDivisor();
  float gyro_rate = 1100 / (1.0 + gyro_divisor);
  Serial.print(gyro_rate);
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Gyroscope in Radians");
  Serial.println("X\tY\tZ");
}
void initializeIcm20948IMU() {
  if (!icm20948.begin_I2C()) {
    Serial.println("Failed to initialize ICM20948 IMU");
    while (1);
  }
  // Initialise IMU Device
  icm20948.setAccelRange(ICM20948_ACCEL_RANGE_2_G);
  icm20948.setAccelRateDivisor(4.6);
  Serial.print("ICM20948 Accelerometer sample rate = ");
  uint16_t accel_divisor = icm20948.getAccelRateDivisor();
  float accel_rate = 1125 / (1.0 + accel_divisor);
  Serial.print(accel_rate);
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Acceleration in Meters/Second2");
  Serial.println("X\tY\tZ");

  icm20948.setGyroRateDivisor(4.5);
  Serial.print("ICM20948 Gyroscope Sample Rate = ");
  uint8_t gyro_divisor = icm20948.getGyroRateDivisor();
  float gyro_rate = 1100 / (1.0 + gyro_divisor);
  Serial.println(gyro_rate);
  Serial.println(" Hz");
  Serial.println();
  Serial.println("Gyroscope in Radians");
  Serial.println("X\tY\tZ");
}
void initializeIsm330dhcxIMU() {
  if (!ism330dhcx.begin_I2C()) {
    Serial.println("Failed to initialize ISM330DHCX IMU");
    while (1);
  }
  // Initialise ISM330DHCX IMU Device
  ism330dhcx.setAccelDataRate(LSM6DS_RATE_208_HZ);
  Serial.print("ISM330DHCX Accelerometer Sample Rate = ");
  switch (ism330dhcx.getAccelDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
  }
  Serial.println();
  Serial.println("Acceleration in Meters/Second2");
  Serial.println("X\tY\tZ");

  ism330dhcx.setGyroDataRate(LSM6DS_RATE_208_HZ);
  Serial.print("ISM330DHCX Gyroscope Sample Rate = ");
  switch (ism330dhcx.getGyroDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
  }
  Serial.println();
  Serial.println("Gyroscope in Radians");
  Serial.println("X\tY\tZ");
}
void initializeLsm6dso32IMU() {
  if (!lsm6dso32.begin_I2C()) {
    Serial.println("Failed to initialize LSM6DSO32 IMU");
    while (1);
  }
  // Initialise IMU Device
  lsm6dso32.setAccelDataRate(LSM6DS_RATE_208_HZ);
  Serial.print("LSM6DSO32 Accelerometer Sample Rate = ");
  switch (lsm6dso32.getAccelDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
  }
  Serial.println();
  Serial.println("Acceleration in Meters/Second2");
  Serial.println("X\tY\tZ");

  lsm6dso32.setGyroDataRate(LSM6DS_RATE_208_HZ);
  Serial.print("LSM6DSO32 Gyroscope Sample Rate = ");
  switch (lsm6dso32.getGyroDataRate()) {
    case LSM6DS_RATE_SHUTDOWN:
      Serial.println("0 Hz");
      break;
    case LSM6DS_RATE_12_5_HZ:
      Serial.println("12.5 Hz");
      break;
    case LSM6DS_RATE_26_HZ:
      Serial.println("26 Hz");
      break;
    case LSM6DS_RATE_52_HZ:
      Serial.println("52 Hz");
      break;
    case LSM6DS_RATE_104_HZ:
      Serial.println("104 Hz");
      break;
    case LSM6DS_RATE_208_HZ:
      Serial.println("208 Hz");
      break;
    case LSM6DS_RATE_416_HZ:
      Serial.println("416 Hz");
      break;
    case LSM6DS_RATE_833_HZ:
      Serial.println("833 Hz");
      break;
    case LSM6DS_RATE_1_66K_HZ:
      Serial.println("1.66 KHz");
      break;
    case LSM6DS_RATE_3_33K_HZ:
      Serial.println("3.33 KHz");
      break;
    case LSM6DS_RATE_6_66K_HZ:
      Serial.println("6.66 KHz");
      break;
  }
  Serial.println();
  Serial.println("Gyroscope in Radians");
  Serial.println("X\tY\tZ");
}
//
// End of IMU Initialization Functions
//


//
// IMU SensorFusion Calibration Functions
//
void sfBmi270Calibrate() {
  IMU.readAccelGyro(bmi270ImuData[0][0], bmi270ImuData[0][2], bmi270ImuData[0][1], bmi270ImuData[0][3], bmi270ImuData[0][5], bmi270ImuData[0][4]);
  bmi270ImuData[0][0] = bmi270ImuData[0][0] * GRAVITY_METERS_SECOND_2;
  bmi270ImuData[0][1] = bmi270ImuData[0][1] * GRAVITY_METERS_SECOND_2;
  bmi270ImuData[0][2] = bmi270ImuData[0][2] * GRAVITY_METERS_SECOND_2;
  bmi270ImuData[0][0] = (bmi270ImuData[0][0] - bmi270OffsetX) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleX);
  bmi270ImuData[0][1] = (bmi270ImuData[0][1] - bmi270OffsetZ) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleZ);
  bmi270ImuData[0][2] = (bmi270ImuData[0][2] - bmi270OffsetY) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleY);
  bmi270ImuData[0][3] = bmi270ImuData[0][3] * DEG_TO_RAD;
  bmi270ImuData[0][4] = bmi270ImuData[0][4] * DEG_TO_RAD;
  bmi270ImuData[0][5] = bmi270ImuData[0][5] * DEG_TO_RAD;

  bmi270Deltat = bmi270Fusion.deltatUpdate();
  bmi270Fusion.MahonyUpdate(bmi270ImuData[0][3], bmi270ImuData[0][4], bmi270ImuData[0][5], bmi270ImuData[0][0], bmi270ImuData[0][1], bmi270ImuData[0][2], bmi270Deltat);
  quat = bmi270Fusion.getQuat();
}
void sfIcm20649Calibrate() {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  icm20649.getEvent(&accel, &gyro, &temp);
  icm20649ImuData[0][0] = (accel.acceleration.x - icm20649OffsetX) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleX);
  icm20649ImuData[0][1] = (accel.acceleration.y - icm20649OffsetY) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleY);
  icm20649ImuData[0][2] = (accel.acceleration.z - icm20649OffsetZ) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleZ);
  icm20649ImuData[0][3] = gyro.gyro.x * DEG_TO_RAD;
  icm20649ImuData[0][4] = gyro.gyro.y * DEG_TO_RAD;
  icm20649ImuData[0][5] = gyro.gyro.z * DEG_TO_RAD;

  icm20649Deltat = icm20649Fusion.deltatUpdate();
  icm20649Fusion.MahonyUpdate(icm20649ImuData[0][3], icm20649ImuData[0][4], icm20649ImuData[0][5], icm20649ImuData[0][0], icm20649ImuData[0][1], icm20649ImuData[0][2], icm20649Deltat);
  quat = icm20649Fusion.getQuat();
}
void sfIcm20948Calibrate() {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  sensors_event_t mag;
  icm20948.getEvent(&accel, &gyro, &temp, &mag);
  icm20948ImuData[0][0] = (accel.acceleration.x - icm20948OffsetX) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleX);
  icm20948ImuData[0][1] = (accel.acceleration.y - icm20948OffsetY) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleY);
  icm20948ImuData[0][2] = (accel.acceleration.z - icm20948OffsetZ) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleZ);
  icm20948ImuData[0][3] = gyro.gyro.x * DEG_TO_RAD;
  icm20948ImuData[0][4] = gyro.gyro.y * DEG_TO_RAD;
  icm20948ImuData[0][5] = gyro.gyro.z * DEG_TO_RAD;

  icm20948Deltat = icm20948Fusion.deltatUpdate();
  icm20948Fusion.MahonyUpdate(icm20948ImuData[0][3], icm20948ImuData[0][4], icm20948ImuData[0][5], icm20948ImuData[0][0], icm20948ImuData[0][1], icm20948ImuData[0][2], icm20948Deltat);
  quat = icm20948Fusion.getQuat();
}
void sfIsm330dhcxCalibrate() {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  ism330dhcx.getEvent(&accel, &gyro, &temp);
  ism330dhcxImuData[0][0] = (accel.acceleration.x - ism330dhcxOffsetX) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleX);
  ism330dhcxImuData[0][1] = (accel.acceleration.y - ism330dhcxOffsetY) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleY);
  ism330dhcxImuData[0][2] = (accel.acceleration.z - ism330dhcxOffsetZ) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleZ);
  ism330dhcxImuData[0][3] = gyro.gyro.x * DEG_TO_RAD;
  ism330dhcxImuData[0][4] = gyro.gyro.y * DEG_TO_RAD;
  ism330dhcxImuData[0][5] = gyro.gyro.z * DEG_TO_RAD;

  ism330dhcxDeltat = ism330dhcxFusion.deltatUpdate();
  ism330dhcxFusion.MahonyUpdate(ism330dhcxImuData[0][3], ism330dhcxImuData[0][4], ism330dhcxImuData[0][5], ism330dhcxImuData[0][0], ism330dhcxImuData[0][1], ism330dhcxImuData[0][2], ism330dhcxDeltat);
  quat = ism330dhcxFusion.getQuat();
}
void sfLsm6dso32Calibrate() {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  lsm6dso32.getEvent(&accel, &gyro, &temp);
  lsm6dso32ImuData[0][0] = (accel.acceleration.x - lsm6dso32OffsetX) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleX);
  lsm6dso32ImuData[0][1] = (accel.acceleration.y - lsm6dso32OffsetY) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleY);
  lsm6dso32ImuData[0][2] = (accel.acceleration.z - lsm6dso32OffsetZ) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleZ);
  lsm6dso32ImuData[0][3] = gyro.gyro.x * DEG_TO_RAD;
  lsm6dso32ImuData[0][4] = gyro.gyro.y * DEG_TO_RAD;
  lsm6dso32ImuData[0][5] = gyro.gyro.z * DEG_TO_RAD;

  lsm6dso32Deltat = lsm6dso32Fusion.deltatUpdate();
  lsm6dso32Fusion.MahonyUpdate(lsm6dso32ImuData[0][3], lsm6dso32ImuData[0][4], lsm6dso32ImuData[0][5], lsm6dso32ImuData[0][0], lsm6dso32ImuData[0][1], lsm6dso32ImuData[0][2], lsm6dso32Deltat);
  quat = lsm6dso32Fusion.getQuat();
}
//
// End IMU SensorFusion Calibration Functions
//


//
// Record IMU Data Functions
//
void recordBmi270IMUData(int i) {
  IMU.readAccelGyro(bmi270ImuData[i][0], bmi270ImuData[i][2], bmi270ImuData[i][1], bmi270ImuData[i][3], bmi270ImuData[i][5], bmi270ImuData[i][4]);
}
void recordIcm20649IMUData(int i) {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  icm20649.getEvent(&accel, &gyro, &temp);
  icm20649ImuData[i][0] = accel.acceleration.x;
  icm20649ImuData[i][1] = accel.acceleration.y;
  icm20649ImuData[i][2] = accel.acceleration.z;
  icm20649ImuData[i][3] = gyro.gyro.x;
  icm20649ImuData[i][4] = gyro.gyro.y;
  icm20649ImuData[i][5] = gyro.gyro.z;
}
void recordIcm20948IMUData(int i) {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  sensors_event_t mag;
  icm20948.getEvent(&accel, &gyro, &temp, &mag);
  icm20948ImuData[i][0] = accel.acceleration.x;
  icm20948ImuData[i][1] = accel.acceleration.y;
  icm20948ImuData[i][2] = accel.acceleration.z;
  icm20948ImuData[i][3] = gyro.gyro.x;
  icm20948ImuData[i][4] = gyro.gyro.y;
  icm20948ImuData[i][5] = gyro.gyro.z;
}
void recordIsm330dhcxIMUData(int i) {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  ism330dhcx.getEvent(&accel, &gyro, &temp);
  ism330dhcxImuData[i][0] = accel.acceleration.x;
  ism330dhcxImuData[i][1] = accel.acceleration.y;
  ism330dhcxImuData[i][2] = accel.acceleration.z;
  ism330dhcxImuData[i][3] = gyro.gyro.x;
  ism330dhcxImuData[i][4] = gyro.gyro.y;
  ism330dhcxImuData[i][5] = gyro.gyro.z;
}
void recordLsm6dso32IMUData(int i) {
  sensors_event_t accel;
  sensors_event_t gyro;
  sensors_event_t temp;
  lsm6dso32.getEvent(&accel, &gyro, &temp);
  lsm6dso32ImuData[i][0] = accel.acceleration.x;
  lsm6dso32ImuData[i][1] = accel.acceleration.y;
  lsm6dso32ImuData[i][2] = accel.acceleration.z;
  lsm6dso32ImuData[i][3] = gyro.gyro.x;
  lsm6dso32ImuData[i][4] = gyro.gyro.y;
  lsm6dso32ImuData[i][5] = gyro.gyro.z;
}
//
// End of Record IMU Data Functions
//


//
// Apply SensorFusion to Modify Recorded IMU Data Functions
//
void bmi270ProcessData() {
  for (int i = 0; i < 200; i++) {
    bmi270ImuData[i][0] = bmi270ImuData[i][0] * GRAVITY_METERS_SECOND_2;
    bmi270ImuData[i][1] = bmi270ImuData[i][1] * GRAVITY_METERS_SECOND_2;
    bmi270ImuData[i][2] = bmi270ImuData[i][2] * GRAVITY_METERS_SECOND_2;
    bmi270ImuData[i][0] = (bmi270ImuData[i][0] - bmi270OffsetX) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleX);
    bmi270ImuData[i][1] = (bmi270ImuData[i][1] - bmi270OffsetZ) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleZ);
    bmi270ImuData[i][2] = (bmi270ImuData[i][2] - bmi270OffsetY) * (GRAVITY_METERS_SECOND_2 / bmi270ScaleY);
    bmi270ImuData[i][3] = bmi270ImuData[i][3] * DEG_TO_RAD;
    bmi270ImuData[i][4] = bmi270ImuData[i][4] * DEG_TO_RAD;
    bmi270ImuData[i][5] = bmi270ImuData[i][5] * DEG_TO_RAD;

    bmi270Deltat = bmi270Fusion.deltatUpdate();
    bmi270Fusion.MahonyUpdate(bmi270ImuData[i][3], bmi270ImuData[i][4], bmi270ImuData[i][5], bmi270ImuData[i][0], bmi270ImuData[i][1], bmi270ImuData[i][2], bmi270Deltat);
    quat = bmi270Fusion.getQuat();

    g_x = 2.0f * (quat[1] * quat[3] - quat[0] * quat[2]);
    g_y = 2.0f * (quat[0] * quat[1] + quat[2] * quat[3]);
    g_z = quat[0] * quat[0] - quat[1] * quat[1] - quat[2] * quat[2] + quat[3] * quat[3];

    // Convert g's to meters/second2
    g_x = g_x * GRAVITY_METERS_SECOND_2;
    g_y = g_y * GRAVITY_METERS_SECOND_2;
    g_z = g_z * GRAVITY_METERS_SECOND_2;

    // Remove gravity component
    bmi270ImuData[i][6] = bmi270ImuData[i][0] - g_x;
    bmi270ImuData[i][7] = bmi270ImuData[i][1] - g_y;
    bmi270ImuData[i][8] = bmi270ImuData[i][2] - g_z;
    bmi270ImuData[i][9] = sqrt((bmi270ImuData[i][6] * bmi270ImuData[i][6]) + (bmi270ImuData[i][7] * bmi270ImuData[i][7]) + (bmi270ImuData[i][8] * bmi270ImuData[i][8]));
  }
}
void icm20649ProcessData() {
  for (int i = 0; i < 200; i++) {
    icm20649ImuData[i][0] = (icm20649ImuData[i][0] - icm20649OffsetX) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleX);
    icm20649ImuData[i][1] = (icm20649ImuData[i][1] - icm20649OffsetY) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleY);
    icm20649ImuData[i][2] = (icm20649ImuData[i][2] - icm20649OffsetZ) * (GRAVITY_METERS_SECOND_2 / icm20649ScaleZ);
    icm20649ImuData[i][3] = icm20649ImuData[i][3] * DEG_TO_RAD;
    icm20649ImuData[i][4] = icm20649ImuData[i][4] * DEG_TO_RAD;
    icm20649ImuData[i][5] = icm20649ImuData[i][5] * DEG_TO_RAD;

    icm20649Deltat = icm20649Fusion.deltatUpdate();
    icm20649Fusion.MahonyUpdate(icm20649ImuData[i][3], icm20649ImuData[i][4], icm20649ImuData[i][5], icm20649ImuData[i][0], icm20649ImuData[i][1], icm20649ImuData[i][2], icm20649Deltat);
    quat = icm20649Fusion.getQuat();

    g_x = 2.0f * (quat[1] * quat[3] - quat[0] * quat[2]);
    g_y = 2.0f * (quat[0] * quat[1] + quat[2] * quat[3]);
    g_z = quat[0] * quat[0] - quat[1] * quat[1] - quat[2] * quat[2] + quat[3] * quat[3];

    // Convert g's to meters/second2
    g_x = g_x * GRAVITY_METERS_SECOND_2;
    g_y = g_y * GRAVITY_METERS_SECOND_2;
    g_z = g_z * GRAVITY_METERS_SECOND_2;

    // Remove gravity component
    icm20649ImuData[i][6] = icm20649ImuData[i][0] - g_x;
    icm20649ImuData[i][7] = icm20649ImuData[i][1] - g_y;
    icm20649ImuData[i][8] = icm20649ImuData[i][2] - g_z;
    icm20649ImuData[i][9] = sqrt((icm20649ImuData[i][6] * icm20649ImuData[i][6]) + (icm20649ImuData[i][7] * icm20649ImuData[i][7]) + (icm20649ImuData[i][8] * icm20649ImuData[i][8]));
  }
}
void icm20948ProcessData() {
  for (int i = 0; i < 200; i++) {
    icm20948ImuData[i][0] = (icm20948ImuData[i][0] - icm20948OffsetX) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleX);
    icm20948ImuData[i][1] = (icm20948ImuData[i][1] - icm20948OffsetY) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleY);
    icm20948ImuData[i][2] = (icm20948ImuData[i][2] - icm20948OffsetZ) * (GRAVITY_METERS_SECOND_2 / icm20948ScaleZ);
    icm20948ImuData[i][3] = icm20948ImuData[i][3] * DEG_TO_RAD;
    icm20948ImuData[i][4] = icm20948ImuData[i][4] * DEG_TO_RAD;
    icm20948ImuData[i][5] = icm20948ImuData[i][5] * DEG_TO_RAD;

    icm20948Deltat = icm20948Fusion.deltatUpdate();
    icm20948Fusion.MahonyUpdate(icm20948ImuData[i][3], icm20948ImuData[i][4], icm20948ImuData[i][5], icm20948ImuData[i][0], icm20948ImuData[i][1], icm20948ImuData[i][2], icm20948Deltat);
    quat = icm20948Fusion.getQuat();

    g_x = 2.0f * (quat[1] * quat[3] - quat[0] * quat[2]);
    g_y = 2.0f * (quat[0] * quat[1] + quat[2] * quat[3]);
    g_z = quat[0] * quat[0] - quat[1] * quat[1] - quat[2] * quat[2] + quat[3] * quat[3];

    // Convert g's to meters/second2
    g_x = g_x * GRAVITY_METERS_SECOND_2;
    g_y = g_y * GRAVITY_METERS_SECOND_2;
    g_z = g_z * GRAVITY_METERS_SECOND_2;

    // Remove gravity component
    icm20948ImuData[i][6] = icm20948ImuData[i][0] - g_x;
    icm20948ImuData[i][7] = icm20948ImuData[i][1] - g_y;
    icm20948ImuData[i][8] = icm20948ImuData[i][2] - g_z;
    icm20948ImuData[i][9] = sqrt((icm20948ImuData[i][6] * icm20948ImuData[i][6]) + (icm20948ImuData[i][7] * icm20948ImuData[i][7]) + (icm20948ImuData[i][8] * icm20948ImuData[i][8]));
  }
}
void ism330dhcxProcessData() {
  for (int i = 0; i < 200; i++) {
    ism330dhcxImuData[i][0] = (ism330dhcxImuData[i][0] - ism330dhcxOffsetX) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleX);
    ism330dhcxImuData[i][1] = (ism330dhcxImuData[i][1] - ism330dhcxOffsetY) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleY);
    ism330dhcxImuData[i][2] = (ism330dhcxImuData[i][2] - ism330dhcxOffsetZ) * (GRAVITY_METERS_SECOND_2 / ism330dhcxScaleZ);
    ism330dhcxImuData[i][3] = ism330dhcxImuData[i][3] * DEG_TO_RAD;
    ism330dhcxImuData[i][4] = ism330dhcxImuData[i][4] * DEG_TO_RAD;
    ism330dhcxImuData[i][5] = ism330dhcxImuData[i][5] * DEG_TO_RAD;

    ism330dhcxDeltat = ism330dhcxFusion.deltatUpdate();
    ism330dhcxFusion.MahonyUpdate(ism330dhcxImuData[i][3], ism330dhcxImuData[i][4], ism330dhcxImuData[i][5], ism330dhcxImuData[i][0], ism330dhcxImuData[i][1], ism330dhcxImuData[i][2], ism330dhcxDeltat);
    quat = ism330dhcxFusion.getQuat();

    g_x = 2.0f * (quat[1] * quat[3] - quat[0] * quat[2]);
    g_y = 2.0f * (quat[0] * quat[1] + quat[2] * quat[3]);
    g_z = quat[0] * quat[0] - quat[1] * quat[1] - quat[2] * quat[2] + quat[3] * quat[3];

    // Convert g's to meters/second2
    g_x = g_x * GRAVITY_METERS_SECOND_2;
    g_y = g_y * GRAVITY_METERS_SECOND_2;
    g_z = g_z * GRAVITY_METERS_SECOND_2;

    // Remove gravity component
    ism330dhcxImuData[i][6] = ism330dhcxImuData[i][0] - g_x;
    ism330dhcxImuData[i][7] = ism330dhcxImuData[i][1] - g_y;
    ism330dhcxImuData[i][8] = ism330dhcxImuData[i][2] - g_z;
    ism330dhcxImuData[i][9] = sqrt((ism330dhcxImuData[i][6] * ism330dhcxImuData[i][6]) + (ism330dhcxImuData[i][7] * ism330dhcxImuData[i][7]) + (ism330dhcxImuData[i][8] * ism330dhcxImuData[i][8]));
  }
}
void lsm6dso32ProcessData() {
  for (int i = 0; i < 200; i++) {
    lsm6dso32ImuData[i][0] = (lsm6dso32ImuData[i][0] - lsm6dso32OffsetX) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleX);
    lsm6dso32ImuData[i][1] = (lsm6dso32ImuData[i][1] - lsm6dso32OffsetY) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleY);
    lsm6dso32ImuData[i][2] = (lsm6dso32ImuData[i][2] - lsm6dso32OffsetZ) * (GRAVITY_METERS_SECOND_2 / lsm6dso32ScaleZ);
    lsm6dso32ImuData[i][3] = lsm6dso32ImuData[i][3] * DEG_TO_RAD;
    lsm6dso32ImuData[i][4] = lsm6dso32ImuData[i][4] * DEG_TO_RAD;
    lsm6dso32ImuData[i][5] = lsm6dso32ImuData[i][5] * DEG_TO_RAD;

    lsm6dso32Deltat = lsm6dso32Fusion.deltatUpdate();
    lsm6dso32Fusion.MahonyUpdate(lsm6dso32ImuData[i][3], lsm6dso32ImuData[i][4], lsm6dso32ImuData[i][5], lsm6dso32ImuData[i][0], lsm6dso32ImuData[i][1], lsm6dso32ImuData[i][2], lsm6dso32Deltat);
    quat = lsm6dso32Fusion.getQuat();

    g_x = 2.0f * (quat[1] * quat[3] - quat[0] * quat[2]);
    g_y = 2.0f * (quat[0] * quat[1] + quat[2] * quat[3]);
    g_z = quat[0] * quat[0] - quat[1] * quat[1] - quat[2] * quat[2] + quat[3] * quat[3];

    // Convert g's to meters/second2
    g_x = g_x * GRAVITY_METERS_SECOND_2;
    g_y = g_y * GRAVITY_METERS_SECOND_2;
    g_z = g_z * GRAVITY_METERS_SECOND_2;

    // Remove gravity component
    lsm6dso32ImuData[i][6] = lsm6dso32ImuData[i][0] - g_x;
    lsm6dso32ImuData[i][7] = lsm6dso32ImuData[i][1] - g_y;
    lsm6dso32ImuData[i][8] = lsm6dso32ImuData[i][2] - g_z;
    lsm6dso32ImuData[i][9] = sqrt((lsm6dso32ImuData[i][6] * lsm6dso32ImuData[i][6]) + (lsm6dso32ImuData[i][7] * lsm6dso32ImuData[i][7]) + (lsm6dso32ImuData[i][8] * lsm6dso32ImuData[i][8]));
  }
}
//
// End of SensorFusion to Modify Recorded IMU Data Functions
//


//
// Convert IMU Snapshot Data to Strings for Transfer Over Bluetooth
//
void bmi270SnapshotToString(int i) {
  switch (i) {
    case 0:
      imuSnapshot[i] = "BMI270,,,,,,,,,";
      break;
    case 1:
      imuSnapshot[i] = "aX,aY,aZ,gX,gY,gZ,LaX,LaY,LaZ,LinearAccel Magnitude";
      break;
    default:
      imuSnapshot[i] = String(bmi270ImuData[i-2][0]) + ',' +
              String(bmi270ImuData[i-2][1]) + ',' +
              String(bmi270ImuData[i-2][2]) + ',' +
              String(bmi270ImuData[i-2][3]) + ',' +
              String(bmi270ImuData[i-2][4]) + ',' +
              String(bmi270ImuData[i-2][5]) + ',' +
              String(bmi270ImuData[i-2][6]) + ',' +
              String(bmi270ImuData[i-2][7]) + ',' +
              String(bmi270ImuData[i-2][8]) + ',' +
              String(bmi270ImuData[i-2][9]);
      break;
  }
}
void icm20649SnapshotToString(int i) {
  if (useBmi270) {
    imuSnapshot[i] += ',';
  }
  switch (i) {
    case 0:
      imuSnapshot[i] += "ICM20649,,,,,,,,,";
      break;
    case 1:
      imuSnapshot[i] += "aX,aY,aZ,gX,gY,gZ,LaX,LaY,LaZ,LinearAccel Magnitude";
      break;
    default:
      imuSnapshot[i] += String(icm20649ImuData[i-2][0]) + ',' +
              String(icm20649ImuData[i-2][1]) + ',' +
              String(icm20649ImuData[i-2][2]) + ',' +
              String(icm20649ImuData[i-2][3]) + ',' +
              String(icm20649ImuData[i-2][4]) + ',' +
              String(icm20649ImuData[i-2][5]) + ',' +
              String(icm20649ImuData[i-2][6]) + ',' +
              String(icm20649ImuData[i-2][7]) + ',' +
              String(icm20649ImuData[i-2][8]) + ',' +
              String(icm20649ImuData[i-2][9]);
      break;
  }
}
void icm20948SnapshotToString(int i) {
  if (useBmi270 || useIcm20649) {
    imuSnapshot[i] += ',';
  }
  switch (i) {
    case 0:
      imuSnapshot[i] += "ICM20948,,,,,,,,,";
      break;
    case 1:
      imuSnapshot[i] += "aX,aY,aZ,gX,gY,gZ,LaX,LaY,LaZ,LinearAccel Magnitude";
      break;
    default:
      imuSnapshot[i] += String(icm20948ImuData[i-2][0]) + ',' +
              String(icm20948ImuData[i-2][1]) + ',' +
              String(icm20948ImuData[i-2][2]) + ',' +
              String(icm20948ImuData[i-2][3]) + ',' +
              String(icm20948ImuData[i-2][4]) + ',' +
              String(icm20948ImuData[i-2][5]) + ',' +
              String(icm20948ImuData[i-2][6]) + ',' +
              String(icm20948ImuData[i-2][7]) + ',' +
              String(icm20948ImuData[i-2][8]) + ',' +
              String(icm20948ImuData[i-2][9]);
      break;
  }
}
void ism330dhcxSnapshotToString(int i) {
  if (useBmi270 || useIcm20649 || useIcm20948) {
    imuSnapshot[i] += ',';
  }
  switch (i) {
    case 0:
      imuSnapshot[i] += "ISM330DHCX,,,,,,,,,";
      break;
    case 1:
      imuSnapshot[i] += "aX,aY,aZ,gX,gY,gZ,LaX,LaY,LaZ,LinearAccel Magnitude";
      break;
    default:
      imuSnapshot[i] += String(ism330dhcxImuData[i-2][0]) + ',' +
              String(ism330dhcxImuData[i-2][1]) + ',' +
              String(ism330dhcxImuData[i-2][2]) + ',' +
              String(ism330dhcxImuData[i-2][3]) + ',' +
              String(ism330dhcxImuData[i-2][4]) + ',' +
              String(ism330dhcxImuData[i-2][5]) + ',' +
              String(ism330dhcxImuData[i-2][6]) + ',' +
              String(ism330dhcxImuData[i-2][7]) + ',' +
              String(ism330dhcxImuData[i-2][8]) + ',' +
              String(ism330dhcxImuData[i-2][9]);
      break;
  }
}
void lsm6dso32SnapshotToString(int i) {
  if (useBmi270 || useIcm20649 || useIcm20948 || useIsm330dhcx) {
    imuSnapshot[i] += ',';
  }
  switch (i) {
    case 0:
      imuSnapshot[i] += "LSM6DSO32,,,,,,,,,";
      break;
    case 1:
      imuSnapshot[i] += "aX,aY,aZ,gX,gY,gZ,LaX,LaY,LaZ,LinearAccel Magnitude";
      break;
    default:
      imuSnapshot[i] += String(lsm6dso32ImuData[i-2][0]) + ',' +
              String(lsm6dso32ImuData[i-2][1]) + ',' +
              String(lsm6dso32ImuData[i-2][2]) + ',' +
              String(lsm6dso32ImuData[i-2][3]) + ',' +
              String(lsm6dso32ImuData[i-2][4]) + ',' +
              String(lsm6dso32ImuData[i-2][5]) + ',' +
              String(lsm6dso32ImuData[i-2][6]) + ',' +
              String(lsm6dso32ImuData[i-2][7]) + ',' +
              String(lsm6dso32ImuData[i-2][8]) + ',' +
              String(lsm6dso32ImuData[i-2][9]);
      break;
  }
}
//
// End of Convert IMU Snapshot Data to Strings Functions
//


void transferIMUData() {
  delay(1000);
  // Send IMU Data
  for (int i = 0; i < 202; i++) {
    transferIMUDataCharacteristic.writeValue(imuSnapshot[i].c_str());
    Serial.println(imuSnapshot[i]);
  }
  if (useBmi270) {
    transferIMUDataCharacteristic.writeValue(String(BMI270_SAMPLERATE_DELAY_MS));
    Serial.println(BMI270_SAMPLERATE_DELAY_MS);
  } else {
    transferIMUDataCharacteristic.writeValue(String(SAMPLERATE_DELAY_MS));
    Serial.println(SAMPLERATE_DELAY_MS);
  }
  
}